#include <stdio.h>
#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"
#include "driver/uart.h"
#include "esp_log.h"
#include "driver/gpio.h"
#include "Uart.h"



const char *tag = "Uart";

#define BUF_SIZE (1024)
#define RD_BUF_SIZE (BUF_SIZE)

QueueHandle_t uart_queue;


static void uart_event_task(void *pvParameters)
{
    uart_event_t event;
    size_t buff_size;
    uint8_t* dtmp = (uint8_t*) malloc(RD_BUF_SIZE);

    for(;;)
    {
        if(xQueueReceive(uart_queue, &event,portMAX_DELAY)) 
        {
            // bzero(dtmp, RD_BUF_SIZE);
            ESP_LOGI(tag, "uart[%d] event:",  UART_NUM_1);
            switch (event.type)
            {
            case UART_DATA:
                ESP_LOGI(tag, "UART DATA : %d," , event.size);
                uart_read_bytes( UART_NUM_1, dtmp, event.size, portMAX_DELAY);

                // uart_read_bytes(UART_NUM_1, dtmp, event.size, portMAX_DELAY);
                for (int buf_i = 0; buf_i < event.size; buf_i++)
                {
                    ESP_LOGE("tag", "received: %x\n", dtmp[buf_i]);
                }
                memset(dtmp, 0, 1024);
                break;
            case UART_FIFO_OVF:
                ESP_LOGI(tag, "fifo overflow");
                uart_flush_input( UART_NUM_1);
                xQueueReset( uart_queue);
                break;
            case UART_BUFFER_FULL:
                ESP_LOGI(tag, "ring buffer full");
                // If buffer full happened, you should consider encreasing your buffer size
                // As an example, we directly flush the rx buffer here in order to read more data.
                uart_flush_input( UART_NUM_1);
                xQueueReset(uart_queue);
                break;
                //Event of UART RX break detected
            case UART_BREAK:
                ESP_LOGI(tag, "uart rx break");
                break;
            //Event of UART parity check error
            case UART_PARITY_ERR:
                ESP_LOGI(tag, "uart parity error");
                break;
            //Event of UART frame error
            case UART_FRAME_ERR:
                ESP_LOGI(tag, "uart frame error");
                break;
            default:
                break;
            }
        }
    }

}

void uartInit()
{
    const uart_config_t uart = {
        .baud_rate = 115200,
        .data_bits = UART_DATA_8_BITS,
        .parity = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE,
        .source_clk = UART_SCLK_APB,
    };

    uart_driver_install(UART_NUM_1, BUF_SIZE,BUF_SIZE, 20, &uart_queue,0);
    uart_param_config(UART_NUM_1, &uart);
    uart_set_pin(UART_NUM_1, GPIO_NUM_17, GPIO_NUM_16, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE);
    xTaskCreate(uart_event_task, "uart_event_task", 512, NULL, 12, NULL);
}

