/*******************************************************************************
* Copyright 2011-2012, Cypress Semiconductor Corporation.  All rights reserved.
* You may use this file only in accordance with the license, terms, conditions, 
* disclaimers, and limitations in the end user license agreement accompanying 
* the software package with which this file was provided.
********************************************************************************/

#include <stdlib.h>
#include "cybtldr_command.h"
#include "cybtldr_api.h"
#include "driver/uart.h"
#include "Uart.h"

/* The highest number of memory arrays for any device. This includes flash and EEPROM arrays */
#define MAX_DEV_ARRAYS    0x80
/* The default value if a flash array has not yet received data */
#define NO_FLASH_ARRAY_DATA 0
/* The maximum number of flash arrays */
#define MAX_FLASH_ARRAYS 0x40
/* The minimum array id for EEPROM arrays. */
#define MIN_EEPROM_ARRAY 0x40

uint32_t g_validRows[MAX_FLASH_ARRAYS];

static uint16_t min_uint16(uint16_t a, uint16_t b) 
{
    return (a < b) ? a : b;
}

int CyBtldr_TransferData(uint8_t* inBuf, int inSize, uint8_t* outBuf, int outSize)
{
    uint8_t TestBuff_Sajag[25];
       int err = UART_WriteData(inBuf, inSize);
    printf("HI STACK OVERFLOW\n");
       if (CYRET_SUCCESS == err)
       printf("HI STACK OVERFLOW 2\n");
       err = UART_ReadData(&TestBuff_Sajag, outSize);

    if (CYRET_SUCCESS != err)
        err |= CYRET_ERR_COMM_MASK;

    return err;
}

int CyBtldr_ValidateRow(uint8_t arrayId, uint16_t rowNum)
{
    uint32_t inSize;
    uint32_t outSize;
    uint16_t minRow = 0;
    uint16_t maxRow = 0;
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint8_t status = CYRET_SUCCESS;
    int err = CYRET_SUCCESS;

    if (arrayId < MAX_FLASH_ARRAYS)
    {
        if (NO_FLASH_ARRAY_DATA == g_validRows[arrayId])
        {
            err = CyBtldr_CreateGetFlashSizeCmd(arrayId, inBuf, &inSize, &outSize);
            if (CYRET_SUCCESS == err)
                err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
            if (CYRET_SUCCESS == err)
                err = CyBtldr_ParseGetFlashSizeCmdResult(outBuf, outSize, &minRow, &maxRow, &status);
            if (CYRET_SUCCESS != status)
                err = status | CYRET_ERR_BTLDR_MASK;

            if (CYRET_SUCCESS == err)
            {
                if (CYRET_SUCCESS == status)
                    g_validRows[arrayId] = (minRow << 16) + maxRow;
                else
                    err = status | CYRET_ERR_BTLDR_MASK;
            }
        }
        if (CYRET_SUCCESS == err)
        {
            minRow = (uint16_t)(g_validRows[arrayId] >> 16);
            maxRow = (uint16_t)g_validRows[arrayId];
            if (rowNum < minRow || rowNum > maxRow)
                err = CYRET_ERR_ROW;
        }
    }
    else
        err = CYRET_ERR_ARRAY;

    return err;
}

int CyBtldr_StartBootloadOperation( uint32_t expSiId,
            uint8_t expSiRev, uint32_t* blVer, const uint8_t* securityKeyBuf)
{
    const uint32_t SUPPORTED_BOOTLOADER = 0x010000;
    const uint32_t BOOTLOADER_VERSION_MASK = 0xFF0000;
    uint32_t i;
    uint8_t inBuf[512];
    uint8_t outBuf[512];
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint32_t siliconId = 0;

    uint8_t siliconRev = 0;
    uint8_t status = CYRET_SUCCESS;
    int err;

    for (i = 0; i < MAX_FLASH_ARRAYS; i++)
        g_validRows[i] = NO_FLASH_ARRAY_DATA;

    err = CYRET_SUCCESS;
    if (CYRET_SUCCESS != err)
        err |= CYRET_ERR_COMM_MASK;

    if (CYRET_SUCCESS == err){
        err = CyBtldr_CreateEnterBootLoaderCmd(inBuf, &inSize, &outSize, securityKeyBuf);
    }
        
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseEnterBootLoaderCmdResult(outBuf, outSize, &siliconId, &siliconRev, blVer, &status);
    else if (CyBtldr_TryParseParketStatus(outBuf, outSize, &status) == CYRET_SUCCESS)
        err = status | CYRET_ERR_BTLDR_MASK; //if the response we get back is a valid packet override the err with the response's status

printf("We sent a NULL and err & status: %d %d \n", err, status);
    if (CYRET_SUCCESS == err)
    {

        if (CYRET_SUCCESS != status){
            
            err = status | CYRET_ERR_BTLDR_MASK;
        }
        if (expSiId != siliconId || expSiRev != siliconRev){
            err = CYRET_ERR_DEVICE;
        }
        else if ((*blVer & BOOTLOADER_VERSION_MASK) != SUPPORTED_BOOTLOADER)
            err = CYRET_ERR_VERSION;
    }
    printf("Returning from Bootloader Operatoin %d\n", err);
    return err;
}

int CyBtldr_GetApplicationStatus(uint8_t appID, uint8_t* isValid, uint8_t* isActive)
{
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint8_t status = CYRET_SUCCESS;
    int err;

    err = CyBtldr_CreateGetAppStatusCmd(appID, inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseGetAppStatusCmdResult(outBuf, outSize, isValid, isActive, &status);
    else if (CyBtldr_TryParseParketStatus(outBuf, outSize, &status) == CYRET_SUCCESS)
        err = status | CYRET_ERR_BTLDR_MASK; //if the response we get back is a valid packet override the err with the response's status

    if (CYRET_SUCCESS == err)
    {
        if (CYRET_SUCCESS != status)
            err = status | CYRET_ERR_BTLDR_MASK;
    }

    return err;
}

int CyBtldr_SetApplicationStatus(uint8_t appID)
{
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint8_t status = CYRET_SUCCESS;
    int err;

    err = CyBtldr_CreateSetActiveAppCmd(appID, inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseSetActiveAppCmdResult(outBuf, outSize, &status);

    if (CYRET_SUCCESS == err)
    {
        if (CYRET_SUCCESS != status)
            err = status | CYRET_ERR_BTLDR_MASK;
    }

    return err;
}

int CyBtldr_EndBootloadOperation(void)
{
    uint32_t inSize;
    uint32_t outSize;
    uint8_t inBuf[MAX_COMMAND_SIZE];

    int err = CyBtldr_CreateExitBootLoaderCmd(inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
    {

        if (CYRET_SUCCESS == err)
            // err = g_comm->CloseConnection();

        if (CYRET_SUCCESS != err)
            err |= CYRET_ERR_COMM_MASK;
    }

    return err;
}

static int SendData(uint8_t* buf, uint16_t size, uint16_t* offset, uint16_t maxRemainingDataSize, uint8_t* inBuf, uint8_t* outBuf)
{
    uint8_t status = CYRET_SUCCESS;
    uint32_t inSize = 0, outSize = 0;
    // size is the total bytes of data to transfer.
    // offset is the amount of data already transfered.
    // a is maximum amount of data allowed to be left over when this function ends.
    // (we can leave some data for caller (programRow, VerifyRow,...) to send.
    // TRANSFER_HEADER_SIZE is the amount of bytes this command header takes up.
    const uint16_t TRANSFER_HEADER_SIZE = 7;
    uint16_t subBufSize = min_uint16((uint16_t)(TRANSFER_HEADER_SIZE), size);
    int err = CYRET_SUCCESS;
    //Break row into pieces to ensure we don't send too much for the transfer protocol
    while ((CYRET_SUCCESS == err) && ((size - (*offset)) > maxRemainingDataSize))
    {
        err = CyBtldr_CreateSendDataCmd(&buf[*offset], subBufSize, inBuf, &inSize, &outSize);
        if (CYRET_SUCCESS == err)
            err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
        if (CYRET_SUCCESS == err)
            err = CyBtldr_ParseSendDataCmdResult(outBuf, outSize, &status);
        if (CYRET_SUCCESS != status)
            err = status | CYRET_ERR_BTLDR_MASK;

        (*offset) += subBufSize;
    }
    return err;
}

int CyBtldr_ProgramRow(uint8_t arrayID, uint16_t rowNum, uint8_t* buf, uint16_t size)
{
    const int TRANSFER_HEADER_SIZE = 10;

    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint32_t inSize;
    uint32_t outSize;
    uint16_t offset = 0;
    uint16_t subBufSize;
    uint8_t status = CYRET_SUCCESS;
    int err = CYRET_SUCCESS;
    
    if (arrayID < MAX_FLASH_ARRAYS)
        err = CyBtldr_ValidateRow(arrayID, rowNum);

    if (CYRET_SUCCESS == err)
        err = SendData(buf, size, &offset, (uint16_t)(TRANSFER_HEADER_SIZE), inBuf, outBuf);

    if (CYRET_SUCCESS == err)
    {
        subBufSize = (uint16_t)(size - offset);

        err = CyBtldr_CreateProgramRowCmd(arrayID, rowNum, &buf[offset], subBufSize, inBuf, &inSize, &outSize);
        if (CYRET_SUCCESS == err)
            err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
        if (CYRET_SUCCESS == err)
            err = CyBtldr_ParseProgramRowCmdResult(outBuf, outSize, &status);
        if (CYRET_SUCCESS != status)
            err = status | CYRET_ERR_BTLDR_MASK;
    }

    return err;
}

int CyBtldr_EraseRow(uint8_t arrayID, uint16_t rowNum)
{
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t status = CYRET_SUCCESS;
    int err = CYRET_SUCCESS;
    
    if (arrayID < MAX_FLASH_ARRAYS)
        err = CyBtldr_ValidateRow(arrayID, rowNum);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_CreateEraseRowCmd(arrayID, rowNum, inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseEraseRowCmdResult(outBuf, outSize, &status);
    if (CYRET_SUCCESS != status)
        err = status | CYRET_ERR_BTLDR_MASK;

    return err;
}

int CyBtldr_VerifyRow(uint8_t arrayID, uint16_t rowNum, uint8_t checksum)
{
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t rowChecksum = 0;
    uint8_t status = CYRET_SUCCESS;
    int err = CYRET_SUCCESS;
    
    if (arrayID < MAX_FLASH_ARRAYS)
        err = CyBtldr_ValidateRow(arrayID, rowNum);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_CreateVerifyRowCmd(arrayID, rowNum, inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseVerifyRowCmdResult(outBuf, outSize, &rowChecksum, &status);
    if (CYRET_SUCCESS != status)
        err = status | CYRET_ERR_BTLDR_MASK;
    if ((CYRET_SUCCESS == err) && (rowChecksum != checksum))
        err = CYRET_ERR_CHECKSUM;

    return err;
}

int CyBtldr_VerifyApplication()
{
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t checksumValid = 0;
    uint8_t status = CYRET_SUCCESS;

    int err = CyBtldr_CreateVerifyChecksumCmd(inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseVerifyChecksumCmdResult(outBuf, outSize, &checksumValid, &status);
    if (CYRET_SUCCESS != status)
        err = status | CYRET_ERR_BTLDR_MASK;
    if ((CYRET_SUCCESS == err) && (!checksumValid))
        err = CYRET_ERR_CHECKSUM;

    return err;
}



int CyBtldr_SetApplicationMetaData(uint8_t appId, uint32_t appStartAddr, uint32_t appSize)
{
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];
    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t status = CYRET_SUCCESS;

    uint8_t metadata[8];
    metadata[0] = (uint8_t)appStartAddr;
    metadata[1] = (uint8_t)(appStartAddr >> 8);
    metadata[2] = (uint8_t)(appStartAddr >> 16);
    metadata[3] = (uint8_t)(appStartAddr >> 24);
    metadata[4] = (uint8_t)appSize;
    metadata[5] = (uint8_t)(appSize >> 8);
    metadata[6] = (uint8_t)(appSize >> 16);
    metadata[7] = (uint8_t)(appSize >> 24);
    int err = CyBtldr_CreateSetApplicationMetadataCmd(appId, metadata, inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseDefaultCmdResult(outBuf, outSize, &status);
    if (CYRET_SUCCESS != status)
        err = status | CYRET_ERR_BTLDR_MASK;

    return err;
}

int CyBtldr_SetEncryptionInitialVector(uint16_t size, uint8_t* buf)
{
    uint8_t inBuf[MAX_COMMAND_SIZE];
    uint8_t outBuf[MAX_COMMAND_SIZE];

    uint32_t inSize = 0;
    uint32_t outSize = 0;
    uint8_t status = CYRET_SUCCESS;

    int err = CyBtldr_CreateSetEncryptionInitialVectorCmd(buf, size, inBuf, &inSize, &outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_TransferData(inBuf, inSize, outBuf, outSize);
    if (CYRET_SUCCESS == err)
        err = CyBtldr_ParseDefaultCmdResult(outBuf, outSize, &status);
    if (CYRET_SUCCESS != status)
        err = status | CYRET_ERR_BTLDR_MASK;

    return err;
}


int UART_ReadData(uint8_t* data, int length) {
    // Wrapper for uart_read_bytes
    printf("Reading Data\n");
    int len = uart_read_bytes(UART_NUM_1, data, length, pdMS_TO_TICKS(1000));
    
    return 0; // Returns the number of bytes read
}

int UART_WriteData(uint8_t* data, int length) {
    // Wrapper for uart_write_bytes
    for (int buf_i = 0; buf_i < length; buf_i++)
    {
        printf("sent: %x\n", data[buf_i]);
    }
    int len = uart_write_bytes(UART_NUM_1,  data, length);
    return 0; // Returns the number of bytes written
}




